/*!
 * onBoard.js @VERSION beta
 * Copyright (c) <Redwood Credit Union>. All rights reserved.
 *
 * THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
 * WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
 * MERCHANTABLITY OR NON-INFRINGEMENT.
 * 
 * example usage var myboards = new OnBoard(
 * [{selector:"#ElementId1",title:"Element1 Description", shape:'rectangle'}, {selector:"#ElementId2", title:"Element 2 Description", subText:"More details about functionality", minRadius:50},{selector:".className", title:"Title of function", subText:"Any additonal information"}],{open:true}
 * );
 */
(function () {
    var testObject = {};

    if (!(Object.setPrototypeOf || testObject.__proto__)) {
        var nativeGetPrototypeOf = Object.getPrototypeOf;

        Object.getPrototypeOf = function (object) {
            if (object.__proto__) {
                return object.__proto__;
            } else {
                return nativeGetPrototypeOf.call(Object, object);
            }
        }
    }
})();

(function () {

    // Define our constructor
    this.OnBoard = function () {

        // Classes
        class OnBoardSlide {
            constructor(options) {
                if (typeof options.selector === 'undefined' || typeof options.title === 'undefined') {
                    return;
                }
                this.selector = options.selector;
                this.title = options.title;
                this.subText = "";
                this.maxRadius = 100;
                this.minRadius = 20;
                this.radius = 20;
                this.xOffset = 0;
                this.yOffset = 0;
                this.shape = 'circle';
                this.beforeSlide = null;
                this.afterSlide = null;
                this.noSpotlight = false;
                this.buttonLocation = 'bottom';
                this.noScroll = false;

                if (!(typeof options.subText === 'undefined')) {
                    this.subText = options.subText;
                }
                if (!(typeof options.maxRadius === 'undefined')) {
                    this.maxRadius = options.maxRadius;
                }
                if (!(typeof options.minRadius === 'undefined')) {
                    this.minRadius = options.minRadius;
                }
                if (!(typeof options.xOffset === 'undefined')) {
                    this.xOffset = options.xOffset;
                }
                if (!(typeof options.yOffset === 'undefined')) {
                    this.yOffset = options.yOffset;
                }
                if (!(typeof options.shape === 'undefined')) {
                    this.shape = options.shape;
                }
                if (!(typeof options.radius === 'undefined')) {
                    // used for rectangle
                    this.radius = options.radius;
                }
                if (!(typeof options.noSpotlight === 'undefined')) {
                    this.noSpotlight = options.noSpotlight;
                }
                if (!(typeof options.buttonLocation === 'undefined')) {
                    this.buttonLocation = options.buttonLocation;
                }
                //callbacks
                if (!(typeof options.beforeSlide === 'undefined')) {
                    this.beforeSlide = options.beforeSlide;
                }
                if (!(typeof options.afterSlide === 'undefined')) {
                    this.afterSlide = options.afterSlide;
                }
                if (!(typeof options.noScroll === 'undefined')) {
                    this.noScroll = options.noScroll;
                }

                var elems = document.querySelectorAll(this.selector);
                if (elems.length == 0) { return; }

                this.element = elems[0]; // just works with 1 element for now

                this.valid = true;
                this.order = 0; // gets set after creation
            }
        }


        // Create global element references
        this.elements = {
            overlay: null,
            container: null,
            text: {
                container: null,
                title: null,
                subText: null
            },
            spotlight: null,
            buttons: {
                container: null, next: null, previous: null, close: null
            }
        }
        this.options = null;
        this.slides = [];

        this.listeners = null;
        this.currentSlide = 0;
        this.previousSlide = -1;
        this.useCallback = true;

        this.classNames = {
            overlay: 'onBoard-overlay',
            container: 'onBoard-container',
            highlight: 'onBoard-highlight',
            textContainer: 'onBoard-textContainer',
            title: 'onBoard-title',
            subText: 'onBoard-subText',
            spotlight: 'onBoard-spotlight',
            buttonContainer: 'onBoard-buttonContainer',
            button: 'onBoard-button',
            buttonNext: 'onBoard-button-next',
            buttonPrevious: 'onBoard-button-previous',
            buttonClose: 'onBoard-button-close',
            buttonContainerTop: 'onBoard-buttonContainer-top',
            buttonContainerMiddle: 'onBoard-buttonContainer-middle',
            buttonContainerBottom: 'onBoard-buttonContainer-bottom',
            buttonContainerTextTop: 'onBoard-buttonContainer-textTop',
            buttonContainerTextBottom: 'onBoard-buttonContainer-textBottom',
            displayNone: 'onBoard-hidden'
        }

        // Define option defaults
        var defaults = {
            buttonNextText: 'Next',
            buttonPreviousText: 'Previous',
            buttonCloseText: 'Close',
            closeButton: true,
            overlay: true,
            open: false,
            // callbacks
            beforeOpen: null,
            afterOpen: null,
            beforeClose: null,
            afterClose: null
        }
        this.options = defaults;

        // Create options by extending defaults with the passed in arugments
        if (arguments.length > 0) {
            // options
            let inputOptions = null;
            // slide inputs
            if (Array.isArray(arguments[0])) {
                inputSlides = arguments[0];
                if (inputSlides.length > 0) {
                    for (var j = 0; j < inputSlides.length; j++) {
                        var onBoardSlide = new OnBoardSlide(inputSlides[j]);
                        if (!(typeof onBoardSlide.valid === 'undefined') && onBoardSlide.valid) {
                            onBoardSlide.order = this.slides.length;
                            this.slides.push(onBoardSlide);
                        }
                    }
                }
            }
            else if (arguments[0] && typeof arguments[0] === "object") {
                inputOptions = arguments[0];
            }
            if (arguments.length > 1 && arguments[1] && typeof arguments[1] === "object") {
                inputOptions = arguments[1];
            }
            if (inputOptions != null) {
                this.options = extendDefaults(defaults, inputOptions);
            }
        }

        // Utility method to extend defaults with user options
        function extendDefaults(source, properties) {
            var property;
            for (property in properties) {
                if (properties.hasOwnProperty(property)) {
                    source[property] = properties[property];
                }
            }
            return source;
        }

        // Public Methods

        OnBoard.prototype.open = function () {
            // open code goes here
            this.openOnBoard();
        }
        OnBoard.prototype.addSlide = function (slide) {
            if (!(typeof slide === 'undefined')) {
                var onBoardSlide = new OnBoardSlide(slide);
                if (!(typeof onBoardSlide.valid === 'undefined') && onBoardSlide.valid) {
                    onBoardSlide.order = this.slides.length;
                    this.slides.push(onBoardSlide);
                }
            }
        }
        OnBoard.prototype.close = function () {
            // open code goes here
            this.closeOnBoard();
        }
        OnBoard.prototype.destroy = function () {
            // remove elements
            this.elements.container.remove();
            this.elements.overlay.remove();
        }

        // Private Methods

        // create elements
        this.createElements = function () {
            this.createContainer();
            this.createTextContainer();
            this.createButtons();
            this.createListeners();
            this.createOverlay();
            this.createSpotlight();
        }

        this.createOverlay = function () {
            // create overlay - contains spotlight
            var ele = document.createElement("DIV");
            let ele_id = randstr('onBoard-overlay-');
            ele.setAttribute("id", ele_id);
            // hide overlay by default
            ele.className = this.classNames.overlay + " " + this.classNames.displayNone;
            document.body.appendChild(ele);
            this.elements.overlay = ele;
        }

        this.createContainer = function () {
            // create cotnainer that includes text and buttons
            var ele = document.createElement("DIV");
            let ele_id = randstr('onBoard-container-');
            ele.setAttribute("id", ele_id);
            // hide by default
            ele.className = this.classNames.container + " " + this.classNames.displayNone;
            document.body.appendChild(ele);
            this.elements.container = ele;
        }

        this.createButtons = function () {
            let containerElem = document.createElement("DIV");
            containerElem.className = this.classNames.buttonContainer;
            this.elements.buttons.container = containerElem;
            this.elements.container.appendChild(containerElem);

            const buttonClass = " " + this.classNames.button;
            // previous button
            this.elements.buttons.previous = this.createButton(this.classNames.buttonPrevious + buttonClass, this.options.buttonPreviousText, containerElem);
            // next button
            this.elements.buttons.next = this.createButton(this.classNames.buttonNext + buttonClass, this.options.buttonNextText, containerElem);
            // close button
            if (true || this.options.closeButton) { // always create, needs it for last slide
                this.elements.buttons.close = this.createButton(this.classNames.buttonClose + buttonClass, this.options.buttonCloseText, containerElem);
            }
        }
        this.createButton = function (className, text, containerElem) {
            var ele = document.createElement("BUTTON");
            let ele_id = randstr('onBoard-button-');
            ele.setAttribute("id", ele_id);
            ele.className = className;
            ele.innerHTML = text;
            containerElem.appendChild(ele);
            return ele;
        }
        this.createTextContainer = function () {
            // add text container
            var cont = document.createElement("DIV");
            cont.className = this.classNames.textContainer;
            this.elements.container.appendChild(cont);
            this.elements.text.container = cont;
            // add title
            var h2 = document.createElement("H2");
            h2.className = this.classNames.title;
            this.elements.text.container.appendChild(h2);
            this.elements.text.title = h2;
            // add subText
            var p = document.createElement("P");
            p.className = this.classNames.subText;
            this.elements.text.container.appendChild(p);
            this.elements.text.subText = p;
        }
        this.createSpotlight = function () {
            // add spotlight
            var sl = document.createElement("B");
            sl.className = this.classNames.spotlight;
            this.elements.overlay.appendChild(sl);
            this.elements.spotlight = sl;
        }

        this.createListeners = function () {
            this.elements.buttons.next.addEventListener("click", this.nextSlide.bind(this));
            this.elements.buttons.previous.addEventListener("click", this.previousSlide.bind(this));
            if (true || this.options.closeButton) {
                this.elements.buttons.close.addEventListener("click", this.closeOnBoard.bind(this));
            }
            window.addEventListener("resize", this.redraw.bind(this), false);
        }

        // Private Actions
        this.openOnBoard = function () {
            if (this.elements.overlay == null) {
                this.createElements();
            }
            if (this.options.beforeOpen != null) { this.options.beforeOpen.call(); }
            this.hideShowElement(this.elements.overlay, true);
            this.hideShowElement(this.elements.container, true);
            this.populateSlide();
            if (this.options.afterOpen != null) { this.options.afterOpen.call(); }
        }
        this.closeOnBoard = function () {
            if (this.options.beforeClose != null) { this.options.beforeClose.call(); }
            this.hideShowElement(this.elements.overlay, false);
            this.hideShowElement(this.elements.container, false);
            if (this.options.afterClose != null) { this.options.afterClose.call(); }
        }
        this.nextSlide = function () {
            this.previousSlide = this.currentSlide;
            this.currentSlide++;
            this.populateSlide();
        }
        this.previousSlide = function () {
            this.previousSlide = this.currentSlide;
            this.currentSlide--;
            this.populateSlide();
        }
        this.redraw = function () {
            this.useCallback = false;
            this.populateSlide();
        }

        //private helpers

        this.populateSlide = function () {

            getPosition = function (currentValue, offsetValue, eleRect) {
                postion = 0;
                if (typeof offsetValue == 'number') {
                    postion = currentValue + parseInt(offsetValue);
                } else if (typeof offsetValue == 'string') {
                    if (offsetValue == 'top') {
                        postion = eleRect.top;
                    } else if (offsetValue == 'bottom') {
                        postion = eleRect.bottom;
                    } else if (offsetValue == 'left') {
                        postion = eleRect.left;
                    } else if (offsetValue == 'right') {
                        postion = eleRect.right;
                    }
                }
                return postion;
            }

            if (this.slides.length > this.currentSlide && this.currentSlide >= 0) {

                const vw = Math.max(document.documentElement.clientWidth || 0, window.innerWidth || 0);
                const vh = Math.max(document.documentElement.clientHeight || 0, window.innerHeight || 0);
                let currentSlide = this.slides[this.currentSlide];

                // fire after callback of previous slide
                if (this.useCallback && this.previousSlide > -1 && this.slides[this.previousSlide].afterSlide != null) {
                    this.slides[this.previousSlide].afterSlide.call();
                }

                // fire before callback of current slide
                if (this.useCallback && currentSlide.beforeSlide != null) {
                    currentSlide.beforeSlide.call();
                }
                this.useCallback = true;

                // scroll to element if out of view
                let rect = currentSlide.element.getBoundingClientRect();
                if (!currentSlide.noScroll) {
                    if (rect.y > (window.scrollY + vh)) {
                        currentSlide.element.scrollIntoView({ block: "end", behavior: "smooth" });
                        rect = currentSlide.element.getBoundingClientRect();
                    } else if (rect.y < (window.scrollY)) {
                        //currentSlide.element.scrollIntoView({ block:"end", behavior: "smooth" });
                        window.scrollBy({ top: (rect.top - rect.height - 60), behavior: 'smooth' })
                        rect = currentSlide.element.getBoundingClientRect();
                    }
                }
                this.elements.text.title.innerHTML = currentSlide.title;
                this.elements.text.subText.innerHTML = currentSlide.subText;

                let spotlightRadius = Math.min(Math.round(rect.width / 2), currentSlide.maxRadius);
                spotlightRadius = Math.max(spotlightRadius, currentSlide.minRadius);

                // get dimensions for spotlight
                let spotlightTop = (rect.top + (rect.height) / 2);
                let spotlightLeft = (rect.left + (rect.width) / 2);
                let spotlightHeight = 2;
                let spotlightWidth = 2;
                let borderRadius = "50%";
                if (currentSlide.shape == 'rectangle') {
                    spotlightRadius = currentSlide.radius;
                    spotlightTop = rect.top;
                    spotlightLeft = rect.left;
                    spotlightHeight = rect.height;
                    spotlightWidth = rect.width;
                    borderRadius = "0%";
                } else {
                    spotlightTop = Math.round(getPosition(spotlightTop, currentSlide.yOffset, rect));
                    spotlightLeft = Math.round(getPosition(spotlightLeft, currentSlide.xOffset, rect));
                }

                // to get values for where the text container should go
                let spotlightVisualBottom = (spotlightTop + spotlightRadius + spotlightHeight) + window.scrollY;
                let spotlightVisualTop = (spotlightTop - spotlightRadius) - window.scrollY;

                let spotlightStyle = "top:" + (Math.round(spotlightTop) + window.scrollY) + "px;"
                    + "left:" + Math.round(spotlightLeft) + "px;"
                    + "height:" + Math.round(spotlightHeight) + "px;"
                    + "width:" + Math.round(spotlightWidth) + "px;"
                    + "box-shadow: 0 0 10px " + spotlightRadius + "px #fff;"
                    + "border-radius:" + borderRadius;
                if (currentSlide.noSpotlight) {
                    spotlightStyle = spotlightStyle + ";display:none;";
                }

                this.elements.spotlight.setAttribute("style", spotlightStyle);

                if (spotlightVisualTop > (vh - spotlightVisualBottom)) {
                    // text above spotlight
                    this.elements.text.container.setAttribute("style", "bottom:" + Math.round(vh - spotlightVisualTop) + "px");
                } else {
                    // text below spotlight
                    this.elements.text.container.setAttribute("style", "top:" + spotlightVisualBottom + "px");
                }

                // position buttons (top, middle, bottom, below the text, above the text)
                if (currentSlide.buttonLocation == 'top') {
                    this.elements.container.appendChild(this.elements.buttons.container);
                    this.elements.buttons.container.className = this.classNames.buttonContainer + " " + this.classNames.buttonContainerTop;
                } else if (currentSlide.buttonLocation == 'middle') {
                    this.elements.container.appendChild(this.elements.buttons.container);
                    this.elements.buttons.container.className = this.classNames.buttonContainer + " " + this.classNames.buttonContainerMiddle;
                } else if (currentSlide.buttonLocation == 'text-bottom') {
                    this.elements.text.container.appendChild(this.elements.buttons.container);
                    this.elements.buttons.container.className = this.classNames.buttonContainer + " " + this.classNames.buttonContainerTextBottom;
                } else if (currentSlide.buttonLocation == 'text-top') {
                    this.elements.text.container.insertBefore(this.elements.buttons.container, this.elements.text.container.firstChild);
                    this.elements.buttons.container.className = this.classNames.buttonContainer + " " + this.classNames.buttonContainerTextTop;
                } else {
                    this.elements.container.appendChild(this.elements.buttons.container);
                    this.elements.buttons.container.className = this.classNames.buttonContainer + " " + this.classNames.buttonContainerBottom;
                }

                this.displayButtons();
            } else {
                this.closeOnBoard();
            }
        }
        this.displayButtons = function () {
            // show next button is not last slide
            this.hideShowElement(this.elements.buttons.next, (this.currentSlide + 1 < this.slides.length));
            // show previous button if not first slide
            this.hideShowElement(this.elements.buttons.previous, (this.currentSlide > 0));
            // show close button if last slide or default is show
            this.hideShowElement(this.elements.buttons.close, (this.options.closeButton || this.currentSlide + 1 == this.slides.length));

        }
        this.hideShowElement = function (ele, show) {
            // remove hidden class
            let elementClass = ele.className;
            elementClass = elementClass.replace(this.classNames.displayNone, "");
            // hidden, add hidden class
            if (!show) {
                elementClass = elementClass + " " + this.classNames.displayNone;
            }
            ele.className = elementClass;
        }

        // random string
        function randstr(prefix) {
            return Math.random().toString(36).replace('0.', prefix || '');
        }


        // handle defaults
        if (this.options.open) {
            this.openOnBoard();
        }

    }

}());